#include <assert.h>
/* 16-bit signed integer dot product
 * Use MMX assist if running on a MMX- or SSE- enabled machine;
 * Use SSE2 assist if running on a SSE2-enabled machine;
 * else use portable C code
 *
 * Copyright 2001 Phil Karn
 * May be used under the terms of the GNU public license
 */
#include <malloc.h>
#include "dotprod.h"

#if defined(MMX) || defined(SSE)
long mmx_dotprod(signed short *a,signed short *b,int cnt);
#elif defined(SSE2)
#include "malloc16.h"
long sse2_dotprod(signed short *a,signed short *b,int cnt);
#endif

/* Create and return a descriptor for use with the dot product function */
void *initdp(signed short coeffs[],int len){
  struct dotprod *dp;
  int j;

  if(len == 0)
    return NULL;

  dp = (struct dotprod *)calloc(1,sizeof(struct dotprod));
  dp->len = len;

#if defined(SSE2)
 {
   int i;

   /* Make 8 copies of coefficients, one for each data alignment
    * Note use of special 16-byte-aligned version of calloc()
    */
   for(i=0;i<8;i++){
     dp->coeffs[i] = (signed short *)calloc16Align(1+(len+i-1)/8,
		    8*sizeof(signed short));
     for(j=0;j<len;j++)
       dp->coeffs[i][j+i] = coeffs[j];
   }
 }

#elif defined(SSE) || defined(MMX)
 {
   int i;

   /* Make 4 copies of coefficients, one for each data alignment */
   for(i=0;i<4;i++){
     dp->coeffs[i] = (signed short *)calloc(1+(len+i-1)/4,
		    4*sizeof(signed short));
     for(j=0;j<len;j++)
       dp->coeffs[i][j+i] = coeffs[j];
   }
 }

#else
 /* Just one copy of the coefficients for the C version */
 dp->coeffs = (signed short *)calloc(len,sizeof(signed short));
 for(j=0;j<len;j++)
   dp->coeffs[j] = coeffs[j];
#endif
  return (void *)dp;
}

/* Free a dot product descriptor created earlier */
void freedp(void *p){
  struct dotprod *dp = (struct dotprod *)p;

#if defined(SSE2)
 {
   int i;
   for(i=0;i<8;i++)
     if(dp->coeffs[i])
       free16Align(dp->coeffs[i]);
 }
#elif defined(SSE) || defined(MMX)
 {
   int i;
   for(i=0;i<4;i++)
     if(dp->coeffs[i])
       free(dp->coeffs[i]);
 }
#endif
  free(dp);
}

/* Compute a dot product given a descriptor and an input array
 * The length is taken from the descriptor
 */
#if defined(MMX) || defined(SSE)
/* MMX-assisted version */
long dotprod(void *p,signed short a[]){
  int al;
  signed short *ar;
  struct dotprod *dp = (struct dotprod *)p;
  
  /* Round input data address down to 8 byte boundary
   * NB: depending on the alignment of a[], memory
   * before a[] will be accessed. The contents don't matter since they'll
   * be multiplied by zero coefficients. I can't conceive of any
   * situation where this could cause a segfault since memory protection
   * in the x86 machines is done on much larger boundaries
   */
  ar = (signed short *)((int)a & ~7);
  
  /* Choose one of 4 sets of pre-shifted coefficients. al is both the
   * index into dp->coeffs[] and the number of 0 words padded onto
   * that coefficients array for alignment purposes
   */
  al = a - ar;
  
  /* Call assembler routine to do the work, passing number of 4-word blocks */
  assert(((int)ar & 7) == 0);
  assert(((int)dp->coeffs[al] & 7) == 0);
  return mmx_dotprod(ar,dp->coeffs[al],(dp->len+al-1)/4+1);
}

#elif defined(SSE2)
/* SSE2-assisted version */
long dotprod(void *p,signed short a[]){
  int al;
  signed short *ar;
  struct dotprod *dp = (struct dotprod *)p;
  
  ar = (signed short *)((int)a & ~15);
  al = a - ar;
  
  /* Call assembler routine to do the work, passing number of 8-word blocks */
  assert(((int)ar & 15) == 0);
  assert(((int)dp->coeffs[al] & 15) == 0);
  return sse2_dotprod(ar,dp->coeffs[al],(dp->len+al-1)/8+1);
}

#else
/* Portable C version */
long dotprod(void *p,signed short a[]){
  struct dotprod *dp = (struct dotprod *)p;
  long corr = 0;
  int i;

  for(i=0;i<dp->len;i++){
    corr += (long)a[i] * dp->coeffs[i];
  }
  return corr;
}
#endif


