/* Compute the sum of the squares of a vector of signed shorts

 * Two user-callable functions are defined. One returns a 64-bit sum,
 * the other returns 32 bits. The second is a little faster, but can
 * easily overflow if the caller isn't careful.

 * Three flavors of each function are defined:
 *  SSE2-assisted version
 *  MMX-assisted version (also used on SSE)
 *  Portable C version

 * The SSE2 and MMX assist routines both operate on multiples of
 * 8 words; they differ only in their alignment requirements (8 bytes
 * for MMX, 16 bytes for SSE2)

 * Copyright 2001 Phil Karn, KA9Q
 * May be used under the terms of the GNU Public License (GPL)
 */

#if defined(SSE2)
long long sumsq_wq_sse2(signed short *,int);
long sumsq_wd_sse2(signed short *,int);

long long sumsq_wq(signed short *in,int cnt){
  long long sum = 0;

  /* Handle stuff before the next 8-byte boundary */
  while(((int)in & 15) != 0 && cnt != 0){
    sum += (long)in[0] * in[0];
    in++;
    cnt--;
  }
  sum += sumsq_wq_sse2(in,cnt);
  in += cnt & ~7;
  cnt &= 7;

  /* Handle up to 7 trailing words */
  while(cnt != 0){
    sum += (long)in[0] * in[0];
    in++;
    cnt--;
  }
  return sum;
}

/* SSE2 has 64-bit integer adds, so there's no speed advantage to
 * a 32-bit version. So just call the 64-bit function and return the
 * lower 32 bits.
 */
long sumsq_wd(signed short *in,int cnt){
  return sumsq_wq(in,cnt);
}

#elif defined(MMX) || defined(SSE)
long long sumsq_wq_mmx(signed short *,int);
long sumsq_wd_mmx(signed short *,int);

long long sumsq_wq(signed short *in,int cnt){
  long long sum = 0;

  /* Handle stuff before the next 8-byte boundary */
  while(((int)in & 7) != 0 && cnt != 0){
    sum += (long)in[0] * in[0];
    in++;
    cnt--;
  }
  sum += sumsq_wq_mmx(in,cnt);
  in += cnt & ~7;
  cnt &= 7;

  /* Handle up to 7 words at end */
  while(cnt != 0){
    sum += (long)in[0] * in[0];
    in++;
    cnt--;
  }
  return sum;
}

long sumsq_wd(signed short *in,int cnt){
  long sum = 0;
  
  /* Handle stuff before the next 8-byte boundary */
  while(((int)in & 7) != 0 && cnt != 0){
    sum += (long)in[0] * in[0];
    in++;
    cnt--;
  }
  sum += sumsq_wd_mmx(in,cnt);
  in += cnt & ~7;
  cnt &= 7;

  /* Handle up to 7 words at end */
  while(cnt != 0){
    sum += (long)*in * *in;
    in++;
    cnt--;
  }
  return sum;
}


#else
/* Portable C */
long long sumsq_wq(signed short *in,int cnt){
  long long sum = 0;
  int i;

  for(i=0;i<cnt;i++){
    sum += (long)in[i] * in[i];
  }
  return sum;
}

long sumsq_wd(signed short *in,int cnt){
  long sum = 0;
  int i;

  for(i=0;i<cnt;i++){
    sum += (long)in[i] * in[i];
  }
  return sum;
}
#endif
